package ru.dbotthepony.mc.otm.worldgen.placement

import com.mojang.serialization.MapCodec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.core.BlockPos
import net.minecraft.util.RandomSource
import net.minecraft.util.valueproviders.ClampedNormalFloat
import net.minecraft.util.valueproviders.ConstantFloat
import net.minecraft.util.valueproviders.FloatProvider
import net.minecraft.util.valueproviders.IntProvider
import net.minecraft.util.valueproviders.UniformFloat
import net.minecraft.world.level.levelgen.placement.PlacementModifierType
import ru.dbotthepony.mc.otm.core.math.Vector
import ru.dbotthepony.mc.otm.core.math.plus
import ru.dbotthepony.mc.otm.core.math.toBlockPos
import ru.dbotthepony.mc.otm.core.nextDouble
import ru.dbotthepony.mc.otm.data.world.BooleanProvider
import ru.dbotthepony.mc.otm.registry.data.MPlacementModifiers
import java.util.stream.Stream
import kotlin.math.PI
import kotlin.math.cos
import kotlin.math.sin

class WormPlacement(
	parameters: Parameters,
	val wormLength: IntProvider,
	val wormTurnChance: BooleanProvider,
	val wormTurnXZ: FloatProvider,
	val wormTurnXY: FloatProvider,
) : AbstractEnormousPlacement(parameters) {
	override fun getPositions(center: BlockPos, random: RandomSource): Stream<BlockPos> {
		var position = Vector.ZERO
		val maxDistance = wormLength.sample(random)
		// determine initial worm facing angle
		var xzRotation = random.nextDouble(-PI / 2.0, PI / 2.0)
		var xyRotation = random.nextDouble(-PI / 16.0, PI / 16.0)

		var xzSin = sin(xzRotation)
		var xzCos = cos(xzRotation)
		var xySin = sin(xyRotation)

		val positions = ArrayList<BlockPos>()
		var prevPos = position.toBlockPos()
		positions.add(prevPos)

		for (traveledDistance in 0 .. maxDistance) {
			// wormy turn
			if (wormTurnChance.sample(random)) {
				// wormy angle
				// TODO: smooth turning, instead of snapping to new angle make it gradually face new angle
				xzRotation += wormTurnXZ.sample(random)
				xyRotation += wormTurnXY.sample(random)

				xzSin = sin(xzRotation)
				xzCos = cos(xzRotation)
				xySin = sin(xyRotation)
			}

			// advance worm
			position += Vector(xzCos, xySin, xzSin)
			val calc = position.toBlockPos()

			if (calc != prevPos) {
				// if worm made a wurm, add new position to set
				prevPos = calc
				positions.add(calc)
			}
		}

		return positions.stream().map { it + center }
	}

	override fun type(): PlacementModifierType<*> {
		return MPlacementModifiers.WORM_PLACEMENT
	}

	companion object {
		private fun increment(value: Float): Float {
			var i = 1f

			while (true) {
				val enlarge = Float.MIN_VALUE * i
				val sum = value + enlarge

				if (sum != value) {
					return sum
				} else {
					i += 1f
				}
			}
		}

		fun uniformTurnRate(degrees: Float): FloatProvider {
			return uniformTurnRateRadians(degrees * DEGREES_TO_RADIANS)
		}

		fun uniformTurnRateRadians(rad: Float): FloatProvider {
			require(rad >= 0f) { "Provided value must be non-negative, $rad given" }
			if (rad == 0f) return ConstantFloat.ZERO
			return UniformFloat.of(-rad, increment(rad))
		}

		fun normalDistributedTurnRate(deviation: Float): FloatProvider {
			return normalDistributedTurnRateRadians(deviation * DEGREES_TO_RADIANS)
		}

		fun normalDistributedTurnRateRadians(deviation: Float): FloatProvider {
			require(deviation >= 0f) { "Provided value must be non-negative, $deviation given" }
			if (deviation == 0f) return ConstantFloat.ZERO
			return ClampedNormalFloat.of(0f, deviation, -PI.toFloat() * 2f, PI.toFloat() * 2f)
		}

		private const val DEGREES_TO_RADIANS = PI.toFloat() / 180f

		val CODEC: MapCodec<WormPlacement> = RecordCodecBuilder.mapCodec {
			it.group(
				PARAMETERS_CODEC.forGetter(WormPlacement::parameters),
				IntProvider.CODEC.fieldOf("length").forGetter(WormPlacement::wormLength),
				BooleanProvider.CODEC.fieldOf("turn_chance").forGetter(WormPlacement::wormTurnChance),
				FloatProvider.CODEC.fieldOf("turn_xz").forGetter(WormPlacement::wormTurnXZ),
				FloatProvider.CODEC.fieldOf("turn_xy").forGetter(WormPlacement::wormTurnXY),
			).apply(it, ::WormPlacement)
		}
	}
}
